/*
	Copyright 1985, 1986, 1987, 1988, 1989, 1990, 1991 Chris Lewis
		All Rights Reserved

    See the LICENSE file for a full description of restrictions under which
    this software is provided.

    Function:		Displays SFP font information in Postscript.
			(analogous to showfont, but it reads PK/SFP
			files and generates postscript, rather than
			executing on the printer using the built-in
			fonts)
 */

#ifndef	lint
static char SCCSID[] =
    "@(#)pk2ps.c 2.4 Copyright 92/03/12 01:11:48 Chris Lewis";
#endif
#include "defs.h"
#include "pk.h"

int verbose = 0;
char *progname;
extern char *mustmalloc();
long pks_malloc;
long highest = 0;
int fontonly = 0;

/* character names in standard Adobe encoding */
char *adobecharnames[95] =
    {	"space", "exclam", "quotedbl", "numbersign",
	"dollar", "percent", "ampersand", "quoteright",
	"parenleft", "parenright", "asterisk", "plus",
	"comma", "hyphen", "period", "slash",
	"zero", "one", "two", "three", "four", "five",
	"six", "seven", "eight", "nine", "colon", "semicolon",
	"less", "equal", "greater", "question",
	"at", "A", "B", "C", "D", "E", "F", "G",
	"H", "I", "J", "K", "L", "M", "N", "O",
	"P", "Q", "R", "S", "T", "U", "V", "W",
	"X", "Y", "Z", "bracketleft",
	"backslash", "bracketright", "asciicircum", "underscore",
	"quoteleft", "a", "b", "c", "d", "e", "f", "g",
	"h", "i", "j", "k", "l", "m", "n", "o",
	"p", "q", "r", "s", "t", "u", "v", "w",
	"x", "y", "z", "braceleft", "bar", "braceright", "asciitilde"
    };

extern long sfp_hmi;	/* pitch (default HMI) -- width of space */

/*	Dummied out for pk.c */
needchar(a, b)
int a, b; {
    return(1);
}

main(argc, argv)
int argc;
char **argv; {
    int c;
    extern int optind, getopt();
    extern char *optarg;
    register struct pkc *pc;
    register struct pkp *p;
    extern struct ras *pkrast();
    struct pkp *pk_read();
    register int i, len;
    int chcount;
    int saw_space;
    int lx, ly, ux, uy;		/* character bounding box */
    int flx, fly, fux, fuy;	/* font bounding box */
    char *fontname = NULL;

    progname = argv[0];
    while((c = getopt(argc, argv, "fD:n:v")) != EOF)
	switch(c) {
	    case 'D':
#ifdef	DEBUG
		setdebug(optarg, "diagnostics");
		break;
#else
		fprintf(stderr, "%s: debug not supported recompile with DEBUG\n",
		    progname);
		exit(1);
#endif
	    case 'n':
		fontname = optarg;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'f':
		fontonly = 1;
		break;
	    default:
		fprintf(stderr, "Usage: %s pk_files\n", progname);
		exit(1);
	}
    for (; optind < argc; optind++) {
	char *filebuf = mustmalloc(strlen(argv[optind]) + 10);
	register char *cp;

	highest = 0;

	saw_space = 0;

	cp = strrchr(argv[optind], '/');

	if (cp)
	    strcpy(filebuf, cp+1);
	else
	    strcpy(filebuf, argv[optind]);
	cp = strchr(filebuf, '.');
	if (cp)
	    *cp = '\0';

	if (fontname != NULL) filebuf = fontname;

	printf("%%%%BeginFont: %s\n", filebuf);

	if (!fontonly) {
	    printf("/saveobj save def\n");
	}

	p = pk_read(argv[optind]);

	    /* allocate a new dictionary for the font description */
	printf("10 dict dup begin\n");
	    /* this is a Type 3 font */
	printf("/FontType 3 def\n");
	    /* font translation matrix */
	printf("/FontMatrix [.001 0 0 .001 0 0] def\n");
#ifdef NEVER
	printf("/FontInfo 5 dict def\n");
	printf("FontInfo begin\n");
	printf("    /isFixedPitch false def\n");
	printf("end\n");
#endif
	    /* set up the encoding array */
	printf("/Encoding 256 array def\n");
	printf("0 1 255 {Encoding exch /.notdef put} for\n");
	printf("Encoding");
	chcount = 0;
	for (pc = p->pkp_chars; pc; pc = pc->pkc_next) {
	    c = pc->pkc_char;
	    if (c == 32) saw_space++;
	    printf ((chcount % 3) ? " " : "\n    ");
	    if (pc->pkc_next || !saw_space)
		printf("dup ");
	    if (c >= 32 && c <= 126)
		 printf("%d /%s put", c, adobecharnames[c-32]);
	    else printf("%d /ch%d put", c, c);
	    highest = max(highest, c);
	    chcount++;
	}
	if (!saw_space) {
	    printf ((chcount % 3) ? " " : "\n    ");
	    printf ("32 /space put\n");
	} else putchar ('\n');

	if (verbose) {
	    fprintf(stderr, "Emitting %s, %d codes\n", filebuf, chcount);
	}

	if (highest <= 127)
	    highest = 128;
	else
	    highest = 256;

	    /* define the "BuildChar" routine to generate a character */
	printf("/BuildChar\n");
	    /* the "0" will be filled in with a new dictionary below */
	printf("{ 0 begin\n");
	    /* get the "font dictionary" parameter */
	printf("	/char exch def\n");
	    /* get the "character number" parameter */
	printf("	/fontdict exch def\n");
	    /* translate character number into character name from encoding */
	printf("	/charname fontdict /Encoding get char get def\n");
	    /* get character info array */
	printf("	/charinfo fontdict /CharData get charname get def\n");
	    /* get character width from character info array */
	printf("	/wx charinfo 0 get def\n");
	    /* get character bounding box from character info array */
	printf("	/charbbox charinfo 1 4 getinterval def\n");
	    /* arrange to save the character in the font cache */
	printf("	wx 0 charbbox aload pop setcachedevice\n");
	    /* get bit array dimensions for "imagemask".
	     * "true" means to paint where the 1-bits are */
	printf("	charinfo 5 get charinfo 6 get true\n");
	    /* get image mask matrix */
	printf("	fontdict /imagemaskmatrix get\n");
	    /* insert character translation info into image mask matrix */
	printf("	    dup 4 charinfo 7 get put\n");
	printf("	    dup 5 charinfo 8 get put\n");
	    /* get bitmap data for the character */
	printf("	charinfo 9 1 getinterval cvx\n");
	    /* generate the character */
	printf("	imagemask\n");
	printf("    end\n");
	printf("} def %% BuildChar\n");
	printf("\n");
	    /* replace the first element (#0) of "BuildChar"
	     * with a reference to a brand-new, empty dictionary */
	printf("/BuildChar load 0 6 dict put\n");

#define	KDOTS2UNITS(x)	(double) ((1000*(x)) / ((p->pkp_ds / (pow2(20) * POINT)) * 300))

	    /* define the image mask matrix */
	printf("/imagemaskmatrix [%.4f 0 0 -%.4f 0 0] def\n",
	    1.0 / KDOTS2UNITS((double) 1),
	    1.0 / KDOTS2UNITS((double) 1)
	    );

	    /* define the character data (bitmaps) */
	if (!saw_space) chcount++;
	printf("/CharData %d dict def\n", chcount+1);
	printf("CharData begin\n");
	flx = fly = fux = fuy = 0;
	for (pc = p->pkp_chars; pc; pc = pc->pkc_next) {
	    register struct ras *r;
	    c = pc->pkc_char;
	    r = pkrast(pc);
	    if (!r) {
		fprintf(stderr, "%0: no raster image for %02x\n",
		    progname, c);
		continue;
	    }
	    lx = KDOTS2UNITS((double) -pc->pkc_x_off) + 0.5;
	    ly = KDOTS2UNITS((double) pc->pkc_y_off - pc->pkc_height + 1) + 0.5;
	    ux = KDOTS2UNITS((double) -pc->pkc_x_off + pc->pkc_width) + 0.5;
	    uy = KDOTS2UNITS((double) pc->pkc_y_off + 1) + 0.5;
	    if (lx < flx) flx = lx;
	    if (ly < fly) fly = ly;
	    if (ux > fux) fux = ux;
	    if (uy > fuy) fuy = uy;
	    if (c >= 32 && c <= 126)
		 printf ("    /%s ", adobecharnames[c-32]);
	    else printf ("    /ch%d ", c);
	    printf("[ %d %d %d %d %d %ld %ld %.1f %.1f\n\t<",
		(int) (KDOTS2UNITS((double) pc->pkc_dx / pow2(16)) + 0.5),
		lx, ly, ux, uy,
		r->ras_width,
		r->ras_height,
		(double) pc->pkc_x_off - .5,			/* shift x */
		(double) pc->pkc_y_off - .5			/* shift y */
	    );
	    len = r->ras_height * r->ras_bline;
	    for (i = 0; i < len; i++) {
		printf("%02x", r->ras_raster[i]);
		if (i % 30 == 29 && i < len) printf ("\n\t ");
	    }
	    printf(">]\n    def\n");
	    free(r->ras_raster);
	    free(r);
	}
	if (!saw_space)
	    printf("    /space [ %d 0 0 0 0 1 1 0 0 <>] def\n",
		   (int) (KDOTS2UNITS((double) sfp_hmi) + 0.5));
	printf("    /.notdef [ 0 0 0 0 0 1 0 0 0 <>] def\n");
	printf("end %% CharData\n");

	    /* now that we've seen the whole font,
	     * we know the font bounding box values */
	printf("/FontBBox [%d %d %d %d] def\n", flx, fly, fux, fuy);

	    /* set the "unique ID" to a random value.
	     * Adobe guarantees values from 4,000,000 to 4,999,999
	     * will never be assigned to a commercial font. */
	printf("/UniqueID %ld def\n",
		4000000 + ((long) getpid() + (optind << 16)) % 1000000);

	    /* give the font its name */
	printf("end\n");
	printf("/%s exch definefont pop\n", filebuf);
	printf("%%%%EndFont: %s\n", filebuf);

	/* end of font definition */
	if (!fontonly) {

	    printf("36 %d moveto\n", 11 * 72 - 36);
	    printf("/%s findfont 18 scalefont setfont\n", filebuf);
	    printf("(%s) show\n", argv[optind]);
	    printf("36 %d moveto\n", 11 * 72 - 54);
	    printf("/Courier findfont 18 scalefont setfont\n");
	    printf("(%s) show\n", argv[optind]);

#ifdef	NEVER
	    printf("/%s findfont 10 scalefont setfont\n", filebuf);
	    printf("72 500 moveto (ABCDEFGHIJKLMNOPQRSTUVWXYZ) show\n");
	    printf("72 400 moveto (abcdefghijklmnopqrstuvwxyz) show\n");
#endif

#define	XPOS(i)	(((i) % 16) * 30 + 72)
#define	YPOS(i) (680 - ((i) / 16) * 44)
#define	PSIZE	7
#define	PSEP	8

	    printf("/%s findfont 18 scalefont setfont\n", filebuf);
	    for (i = 0; i < highest; i++) {
		printf("%d %d moveto\n", XPOS(i), YPOS(i) + 3 * PSEP + 3);
		printf("(\\%03o) show\n", i);
	    }
	    printf("/Courier findfont %d scalefont setfont\n", PSIZE);
	    for (i = 0; i < highest; i++) {
		if ((i % 16) == 0) {
		    printf("%d %d moveto\n", 36, YPOS(i) + 2 * PSEP);
		    printf("(DEC:) show\n");
		    printf("%d %d moveto\n", 36, YPOS(i) + PSEP);
		    printf("(HEX:) show\n");
		    printf("%d %d moveto\n", 36, YPOS(i));
		    printf("(OCT:) show\n");
		}
		printf("%d %d moveto\n", XPOS(i), YPOS(i) + 2 * PSEP);
		printf("(%3d) show\n", i);
		printf("%d %d moveto\n", XPOS(i), YPOS(i) + PSEP);
		printf("( %02x) show\n", i);
		printf("%d %d moveto\n", XPOS(i), YPOS(i));
		printf("(%03o) show\n", i);
	    }
	    printf("showpage\n");
	    printf("saveobj restore\004");
	}

	pk_destroy(p);
	free(filebuf);
    }
    exit(0);
}
